; the stash is if the agent needs to store whatever, or you could add
; more struct or object slots as need be
(defstruct agent
  (energy 0 :type fixnum)
  (action nil :type function)
  (stash nil))

(defmacro trigger (object amount)
  "returns true for objects whose energy has fallen to zero"
  (let ((prev (gensym)))
    `(let ((,prev ,object))
       (when (plusp ,prev)
         (decf ,object ,amount)
         (when (minusp ,object) (setf ,object 0))
         (zerop ,object)))))

; One might pass in or have a global time value if an agent needs to do
; different things based on that. The callback should either return nil
; (ignored) or can return new agents.
(defun run-agents (agents amount)
  (loop for ent in agents collect
        (when (trigger (agent-energy ent) amount)
          (funcall (agent-action ent) ent))))

; NOTE this deletes agents with zero energy; to stay alive an agent will
; need to increase its energy value, or spawn new agents. In a game one
; might instead have an "alive?" field and then prune objects that have
; died. Whether or not agents should persist depends on how they are
; being used in the system.
(defmacro doagents (agents amount)
  `(let ((new (run-agents ,agents ,amount)))
     (setf ,agents (delete-if
                     (lambda (ent) (zerop (agent-energy ent)))
                     ,agents))
     ; this compilication is support agents that may return a list of
     ; new agents. another way would be to register new agents, and not
     ; to loop over any new agents until the next tick
     (loop for obj in new do
           (etypecase obj
             (agent (push obj ,agents))
             (list (loop for lobj in obj do
                         (when lobj (push lobj ,agents))))
             (null)))))